/****************************************************************************
**  SIONLIB     http://www.fz-juelich.de/jsc/sionlib                       **
*****************************************************************************
**  Copyright (c) 2008-2016                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/*!
 * \file
 */

#ifndef SION_SION_DATATYPES_MPI
#define SION_SION_DATATYPES_MPI

#ifdef MPI_VERSION

#if defined(_SION_LINUX)
#define SION_MPI_INT32 MPI_INT
#define SION_MPI_INT64 MPI_LONG_LONG
#elif defined(_SION_DARWIN)
#define SION_MPI_INT32 MPI_INT
#define SION_MPI_INT64 MPI_LONG_LONG
#elif defined(_SION_AIX)
#define SION_MPI_INT32 MPI_INT
#define SION_MPI_INT64 MPI_LONG_LONG
#elif defined(_SION_BGL)
#define SION_MPI_INT32 MPI_INTEGER4
#define SION_MPI_INT64 MPI_INTEGER8
#elif defined(_SION_BGP)
#define SION_MPI_INT32 MPI_INTEGER4
#define SION_MPI_INT64 MPI_INTEGER8
#elif defined(_SION_BGQ)
#define SION_MPI_INT32 MPI_INTEGER4
#define SION_MPI_INT64 MPI_INTEGER8
#elif defined(_SION_XT)
#define SION_MPI_INT32 MPI_INTEGER4
#define SION_MPI_INT64 MPI_INTEGER8
#endif

#endif

#endif
